import java.util.*;
import static choco.Choco.*;
import choco.cp.model.CPModel;
import choco.cp.solver.CPSolver;
import choco.kernel.model.Model;
import choco.kernel.solver.Solver;
import choco.kernel.solver.ContradictionException;
import choco.kernel.solver.SolverException;
import choco.kernel.solver.constraints.integer.AbstractLargeIntSConstraint;
import choco.kernel.solver.variables.integer.IntDomainVar;
import choco.kernel.memory.IStateInt;
import choco.kernel.memory.trailing.StoredInt;
import choco.kernel.model.variables.integer.IntegerExpressionVariable;
import choco.kernel.model.variables.integer.IntegerVariable;


public class MinGirth extends AbstractLargeIntSConstraint {
    
    private int n;
    private IntDomainVar [][] A;
    private int [][] map;
    private Stack<Integer> vertices;
    private Stack[] edges;
    private int minGirth;

    public MinGirth(Solver s,IntDomainVar[] S,IntDomainVar[][] A,int n,int minGirth) {
	super(S);
	this.A        = A;
	this.n        = n;
	this.minGirth = minGirth;
	vertices      = new Stack<Integer>();
	edges         = new Stack[minGirth];
	map           = new int[n*(n-1)/2][2];
	for (int i=0,k=0;i<n-1;i++)
	    for (int j=i+1;j<n;j++,k++){
	      map[k][0] = i;
	      map[k][1] = j;
	  }
	for (int i=0;i<minGirth;i++) edges[i] = new Stack<IntDomainVar>();
    }

     private void display(){
	System.out.println();
	for (int i=0;i<n;i++){
	    if (i<10) System.out.print(" ");
	    System.out.print(" "+ i +" ");
	      for (int j=0;j<n;j++)
		  if (i==j) System.out.print("0");
		  else if (A[i][j].isInstantiated()) System.out.print(A[i][j].getVal());
		  else System.out.print("?");
	      System.out.println();
	}
	System.out.println();
    }

    public boolean isSatisfied(){
	// System.out.println("isSatisfied()");
	return true;
    }
    
    public void awake() throws ContradictionException {
	for (int i=0;i<vars.length;i++)
	    if (vars[i].isInstantiated()) awakeOnInst(i);
	// System.out.println("awake()");
    }
    
    public void propagate() throws ContradictionException {
	// System.out.println("propagate()");
    }
    
    public void awakeOnInf(int idx) throws ContradictionException {
	// System.out.println("awakeOnInf(" + idx + ")");
    }
    
    public void awakeOnSup(int idx) throws ContradictionException {
	// System.out.println("awakeOnSup(" + idx + ")");
    }
    
    public void awakeOnRem(int idx,int x) throws ContradictionException {
	// System.out.println("awakeOnRem(" + idx + "," + x + ")");
    }
    
    @SuppressWarnings("unchecked")
    private void  process(int lwb,int girth,int edgeCount,int u,int v) throws ContradictionException {
	//System.out.println("process("+ lwb +","+ girth +","+ edgeCount +","+ u +","+ v +") -> "+ vertices);
	if (girth < minGirth){
	    for (int i=lwb;i<n;i++)
		if (i != u && i != v){
		    int count = edgeCount;
		    edges[girth].clear();
		    for (int w : vertices){
			IntDomainVar edge = A[i][w];
			if (!edge.isInstantiated()) 
			    edges[girth].push(edge);
			else count = count + edge.getVal();
		    }
		    if (count >= girth) A[u][v].setVal(0); // forced exception
		    if (count == girth-1)                  // reject uninstantiated edges
			for (int j=0;j<=girth;j++)
			    while (!edges[j].isEmpty()){IntDomainVar e = (IntDomainVar)edges[j].pop(); e.setVal(0);}
		    vertices.push(i);
		    process(i+1,girth+1,count,u,v);
		    vertices.pop();
		}
	}
    }

    public void awakeOnInst(int idx) throws ContradictionException {awakeOnInst(map[idx][0],map[idx][1]);}

    private void awakeOnInst(int u,int v) throws ContradictionException {
	if (A[u][v].getVal() == 1){ // edge selected
	    	vertices.clear(); 
		vertices.push(u);
		vertices.push(v);
		process(0,3,1,u,v);
	}
    }
}
